var _ = require('underscore'),
    mongodb_schema = require('mongodb-schema'), 
    mongodb_ns = require('mongodb-ns'),
    mongoscope_client = require('mongoscope-client');

var QueryBuilder = function(options) {
    this.options = options || {};

    this._namespace = mongodb_ns(this.options.namespace);
    this._mongoscope = mongoscope_client(this.options);
    this._metavars = this.options.metavars || {count: '#count', type: '#type', data: '#data', array: '#array', prob: '#prob'};
    this._schema = null;
    
    this._getSchema();
};

QueryBuilder.prototype._getSchema = function() {
    // get schema early on
    var that = this;

    // TODO: use sampling, but for now go backwards (newest documents first)
    // this._mongoscope.sample(this._namespace.ns, {size: this.options.samples || 1000}, function (err, docs) {
    this._mongoscope.find(this._namespace.database + '._schema', {limit: 1}, function(err, res) {
        if (err) {
            // no _schema collection available, create schema
            that._mongoscope.find(that._namespace.ns, {limit: that.options.samples || 1000, sort: {_id: -1}}, function (err, docs) {
                if (err) throw err;    

                // analyse schema
                mongodb_schema(docs, {flat: true, data: true, metavars: that._metavars}, function(err, schema) {
                    if (err) throw err;    
                    that._schema = schema;
                });
            });
        } else {
            // use _schema collection as schema
            that._schema = res[0];
        }
  
    });

    this.reset();
};


QueryBuilder.prototype._requireSchema = function(callback, args) {
    var _this = this;
    function wait() {
        if (_this._schema !== null) {
            callback.call(_this, args);
        } else {
            setTimeout(wait, 100);
        }
    }
    wait();
};


QueryBuilder.prototype.schema = function(callback) {
    function inner() {
        callback(null, this._schema);
    };
    this._requireSchema(inner);
    return this;
};


QueryBuilder.prototype.namespace = function (namespace) {
    this._namespace = mongodb_ns(namespace);
    this._getSchema();

    return this;
};

QueryBuilder.prototype.reset = function () {
    this._match = null;
    this._slots = null;
    this._limit = null;
    this._sort = null;
    this._group = null;
    this._agg = null;
    this._unwind = null;
    this._nest_group = null;
    this._nest_agg = null;
    this._pick = null;
    this._bin = null;

    this._picked = false;
    this._grouped = false;
    this._nested = false;
    
    return this;
};

QueryBuilder.prototype.match = function(field, value) {

    // delete fields if value falsy
    if (!value && this._match[field]) {
        delete this._match[field];
        // set _match back to null
        if (Object.keys(this._match).length === 0) {
            this._match = null;
        }
    } else {
        if (!(value instanceof Array)) {
            value = [ value ];
        }
        if (!this._match) {
            this._match = {};
        }     
        this._match[field] = value;
    }
    return this;
};

QueryBuilder.prototype.pick = function(name, field) {
    if (!field) {
        delete this._pick[name];
        // set _pick back to null
        if (Object.keys(this._pick).length === 0) {
            this._pick = null;
            this._picked = false;
        }
    } else {
        this._picked = true;
        if (!this._pick) {
            this._pick = {};
        }     
        this._pick[name] = field;
    }
    return this;
};


QueryBuilder.prototype.limit = function(n) {
    if (n) {
        this._limit = n;    
    } else {
        this._limit = null;
    }
    
    return this;
};


QueryBuilder.prototype.sort = function(field, direction) {
    if (!field) {
        this._sort = null;
    } else {
        this._sort = {"field": field, "direction": direction};
    }
    return this;
};


QueryBuilder.prototype.bin = function(field, options) {
    if (!options) {
        this._bin = null;
    } else {
        this._bin = {field: field, options: options};
    }    

    return this;
};


QueryBuilder.prototype._unwinder = function(field) {    
    // get value from arbitrary key
    if (typeof field !== 'string') {
        return;
    }

    // check if field needs to be unwound
    var tokens = field.split('.');

    for (var i = 1; i < tokens.length + 1; i++) {
        var f = tokens.slice(0, i).join('.');
        // check if field is array, then add to list for unwind
        if ((typeof this._schema[f] === 'object') && (this._schema[f][this._metavars.array] === true)) {
            this._unwind = this._unwind || [];
            if (this._unwind.indexOf(f) === -1) {
                this._unwind.push(f);
            }
        }
    }
};

QueryBuilder.prototype.group = function(name, field) {

    if (!field && this._group[name]) {
        delete this._group[name];
        // set _group back to null
        if (Object.keys(this._group).length === 0) {
            this._group = null;
        }
    } else {
        this._grouped = true;
        
        if (this._group && (Object.keys(this._group).length > 0) && (!(name in this._group))) {

            // a different group, activate nesting
            this._nested = true;
            // move _group to _nest_group and _agg to _nest_agg
            this._nest_group = this._group;
            this._group = null;
            this._nest_agg = this._agg;
            this._agg = null;
        }

        if (!(field instanceof Array)) {
            field = [ field ];
        }
        this._group = this._group || {};
        this._group[name] = field;
    }

    return this;
};

QueryBuilder.prototype.agg = function(name, fn, field) {
    if (!fn) {
        delete this._agg[name];
        // set _pick back to null
        if (Object.keys(this._agg).length === 0) {
            this._agg = null;
        }
    } else {
        this._grouped = true;
        var doc = {};
        doc[fn] = field;
        this._agg = this._agg || {};
        this._agg[name] = doc;
    }
    return this;
};

QueryBuilder.prototype._compileMatch = function() {
    // if nothing to match, return null
    if (!this._match || (Object.keys(this._match).length === 0)) {
        return null;
    }

    var match = {};
    for (var field in this._match) {
        if (!this._match.hasOwnProperty(field)) continue;

        if (!(field in this._schema)) {
            throw Error("can't find field " + field + " in schema");
        }
        
        if (this._schema[field][this._metavars.type] === 'category') {
            if (this._match[field].length === 1) {
                // single matches, no $in
                match[field] = this._match[field][0];
            } else {
                // multiple matches, use $in
                match[field] = { $in: this._match[field] };
            }
        } else if (['number', 'date'].indexOf(this._schema[field][this._metavars.type]) !== -1) {
            // assume length is always 2 (min and max)
            var min = this._match[field][0],
                max = this._match[field][1];

            // extra conversion for date types
            if (this._schema[field][this._metavars.type] === 'date') {
                if (min && !(min instanceof Date)) {
                    min = new Date(Date.parse(min));
                }
                if (max && !(max instanceof Date)) {
                    max = new Date(Date.parse(max));
                }
            } 
            match[field] = {};
            if (min) match[field]['$gte'] = min;
            if (max) match[field]['$lte'] = max;

        } else if (this._schema[field][this._metavars.type] === 'boolean') {
            // convert to boolean
            match[field] = !!this._match[field][0];
        }
    }
    match = {$match: match};
    return match;
};


QueryBuilder.prototype._compileBin = function() {

    if (!this._bin || !this._group) { 
        return null;
    }

    // get field from this._group
    var field = this._bin['field'];
    var options = this._bin['options'];
    var type = this._schema[field][this._metavars.type];

    if (!type || type !== 'date') {
        throw Error("can't bin this type. Only date supported.")
    }

    // auto binning
    if (options === 'auto') {
        var data_range = this._schema[field][this._metavars.data];

        if (this._match && (field in this._match)) {
            // use filter range instead of data range
            var filter_range = this._match[field];

            // fill missing value date from schema 
            filter_range[0] = filter_range[0] || data_range.min;
            filter_range[1] = filter_range[1] || data_range.max;

            // convert to dates if string
            filter_range[0] = (typeof filter_range[0] === 'string') ? new Date(Date.parse(filter_range[0])) : filter_range[0];
            filter_range[1] = (typeof filter_range[1] === 'string') ? new Date(Date.parse(filter_range[1])) : filter_range[1];

        } else {
            var filter_range = [data_range.min, data_range.max]
        }
        
        var range_days = (filter_range[1] - filter_range[0]) / 1000 / 60 / 60 / 24;

        if (range_days > 3650) {
            // ~10 years or more
            options = 'year';
        } else if (range_days > 365) {
            // ~12 months or more
            options = 'month';
        } else if (range_days > 35) {
            // ~5 weeks or more
            options = 'week';
        } else {
            options = 'day';
        }
    }

    // pass all project fields through both parts of the bin $projects
    var project1 = this._compileProject();
    for (var key in project1['$project']) {
        project1['$project'][key] = 1;
    }

    // deep copy project1
    var project2 = JSON.parse(JSON.stringify(project1));

    field = field.replace(/\./g, '_');

    if (type === 'date') {
        var dfield = '$' + field;
        
        project1['$project']['h'] = {"$hour" : dfield};
        project1['$project']['m'] = {"$minute" : dfield};
        project1['$project']['s'] = {"$second" : dfield};
        project1['$project']['ms'] = {"$millisecond" : dfield};

        var day2ms = { "$multiply" : [ {"$subtract": ["$d", 2] }, 24, 60, 60, 1000 ] };

        switch(options) {
            case 'day': var day2ms = 0; break;
            case 'week': project1['$project']['d'] = {'$dayOfWeek': dfield}; break;
            case 'month': project1['$project']['d'] = {'$dayOfMonth': dfield}; break;
            case 'year': project1['$project']['d'] = {'$dayOfYear': dfield}; break;
        }

        project2['$project'][field] = {      
            "$subtract" : [      
              dfield, {
                "$add" : [
                  "$ms", 
                  { "$multiply" : [ "$s", 1000 ] },   
                  { "$multiply" : [ "$m", 60, 1000 ] }, 
                  { "$multiply" : [ "$h", 60, 60, 1000 ] },
                  day2ms
                ]      
              }      
            ]      
        };      
        return [project1, project2];
    }
    return null;

};


QueryBuilder.prototype._compileProject = function() {
    if (this._grouped) {

        var group_name = Object.keys(this._group)[0];
        var group_fields = this._group[group_name];

        // project group fields and agg fields
        var project_fields = group_fields.slice(0);
        for (var agg in this._agg) {
            if (!this._agg.hasOwnProperty(agg)) continue;
            project_fields.push( this._agg[agg][Object.keys(this._agg[agg])[0]] );
        };

        // add nested fields
        if (this._nested) {
            var nest_group_name = Object.keys(this._nest_group)[0];
            var nest_group_fields = this._nest_group[nest_group_name];
            project_fields = project_fields.concat(nest_group_fields);
        }

        // filter out non-strings
        project_fields = project_fields.filter(function (f) { return (typeof f === 'string')});

        // build project object
        var project = project_fields.reduce(function (prev, curr) {
            prev[curr.replace(/\./g, '_')] = '$' + curr;
            return prev;
        }, {});
        project = {$project: project};

    } else if (this._picked) {
        var project = {};

        for (name in this._pick) {
            if (!this._pick.hasOwnProperty(name)) continue;
            var field = this._pick[name];

            // unwind if necessary
            this._unwinder.call(this, field);

            // then project
            project[name] = '$' + field;
        }
        project = {$project: project};


    } else {
        var project = null;
    }

    return project;
};


QueryBuilder.prototype._compileUnwind = function() {
    if (this._unwind === null) {
        return null;
    }

    var unwind = this._unwind.map(function (name) {
        return {'$unwind': '$' + name};
    });
    return unwind;
};


QueryBuilder.prototype._compileGroup = function(nested) {

    // switch between nested and not nested
    var this_group = nested ? this._nest_group : this._group; 
    var this_agg = nested ? this._nest_agg : this._agg;

    // if nothing to group, return null
    if (!this_group || (Object.keys(this_group).length === 0)) {
        return null;
    }

    var group_name = Object.keys(this_group)[0];
    var group_fields = this_group[group_name];

    // check if we need to force binning
    // TODO: this only works on one field currently (last one)
    // TODO: nested vs. normal group, currenty nested is preferred
    var field = group_fields[0];
    if (!this._bin && (this._schema[field][this._metavars.type] === 'date')) {
        this.bin(field, 'auto');
        // override sort order 
        this.sort(group_name, 1);
    }



    // if we nest later on, pull nest fields into group _id
    if (this._nested && !nested) {
        var nest_name = Object.keys(this._nest_group)[0];
        var nest_fields = this._nest_group[nest_name];
        
        nest_fields.forEach(function (f) {
            group_fields.push(f);
        });
    }

    // if nested, add real group name prefix
    var real_group_name = Object.keys(this._group)[0];
    var prefix = nested ? real_group_name + '.' : '';

    // build group _id
    if (group_fields.length > 1) {
        // prep for compound group key
        var compound = {};

        // build compound group key
        group_fields.forEach(function (f) {
            f = f.replace(/\./g, '_');
            compound[f] = '$' + prefix + f;
        });
        var group = { _id: compound };
    } else {
        // single group key
        var group = { _id: '$' + prefix + group_fields[0].replace(/\./g, '_') };
    }

    // unwind group field if necessary
    group_fields.forEach( _.bind(this._unwinder, this) );

    // add aggs to group document
    for (var name in this_agg) {
        if (!this_agg.hasOwnProperty(name)) continue;

        var value = this_agg[name][Object.keys(this_agg[name])[0]];
        var fn = Object.keys(this_agg[name])[0];
                        
        // unwind aggs automatically if necessary
        this._unwinder.call(this, value);

        // add aggregations to group
        var agg = {};
        if (typeof value === 'string') {
            value = '$' + value.replace(/\./g, '_');
        }
        agg[fn] = value;
        group[name] = agg;
    }

    // if we nest later on, push documents into values
    if (nested) {
        group['values'] = {'$push': '$$ROOT'};
    }

    group = {$group: group};
    return group;
};


QueryBuilder.prototype._compileRename = function(nested) {
    if (!this._grouped) {
        return null;
    }

    // switch between nested and not nested
    var this_group = nested ? this._nest_group : this._group; 
    var this_agg = nested ? this._nest_agg : this._agg;

    var group_name = Object.keys(this_group)[0];

    // rename _id to the group name
    var rename = Object.keys(this_agg).reduce(function (doc, curr) {
        doc[curr] = '$' + curr;
        return doc;
    }, {});

    rename[group_name] = '$_id';
    rename['_id'] = 0;

    // project values if nested
    if (nested) {
        rename['values'] = '$values';
    }

    rename = {'$project': rename};

    return rename;
};

QueryBuilder.prototype._compileNest = function() {
    if (!this._nested) {
        return null;
    }

    // create nest_group with _compileGroup but switch to nesting
    var nest_group = this._compileGroup(true);
    
    // create nest_project with _compileProject but switch to nesting
    var nest_rename = this._compileRename(true);

    var nest = [ nest_group, nest_rename ];
    return nest;
};

QueryBuilder.prototype._compileSort = function() {
    if (this._sort === null) {
        return null;
    }

    // sort
    var sort = {};
    sort[this._sort['field']] = this._sort['direction'];
    sort = {$sort: sort};
    return sort;
};

QueryBuilder.prototype._compileLimit = function() {
    if (this._limit === null) {
        return null;
    }

    // limit
    var limit = {'$limit': this._limit};
    return limit;
};


QueryBuilder.prototype.pipeline = function() {

    // compile all stages
    // bin after group and group, they can add forced binning
    // unwind last, other stages add to unwind
    var match_stage = this._compileMatch.call(this);
    var project_stage = this._compileProject.call(this);
    var group_stage = this._compileGroup.call(this);
    var rename_stage = this._compileRename.call(this);
    var nest_stage = this._compileNest.call(this);
    var bin_stage = this._compileBin.call(this);
    var sort_stage = this._compileSort.call(this);
    var limit_stage = this._compileLimit.call(this);
    var unwind_stage = this._compileUnwind.call(this);

    // assemble the pipeline
    var pipeline = [];
    pipeline = pipeline
        .concat( 
            match_stage, 
            unwind_stage, 
            project_stage, 
            bin_stage,
            group_stage, 
            rename_stage,
            nest_stage, 
            sort_stage, 
            limit_stage
        )
        .filter( function (f) { 
            return (f !== null) 
        });

    return pipeline;
};

QueryBuilder.prototype.end = function(callback) {

    function inner() {
        // sanity check
        if (this._grouped && this._picked) {
            throw Error("can't use pick and group/agg together.");
        }

        var pipeline = this.pipeline();
        console.log("PIPELINE", JSON.stringify(pipeline, null, '\t'));

        // executing the pipeline
        this._mongoscope.aggregate(this._namespace.ns, pipeline, callback);
    }

    this._requireSchema(inner);
};



module.exports = QueryBuilder;
