var QueryBuilder = require('../'),
          assert = require('assert'),
          schema = require('./schema');

describe('QueryBuilder', function() {
    
    var builder = new QueryBuilder({seed: "mongodb://localhost:10000", namespace: "users.users"});

    beforeEach(function(done) {
        // create query builder for each test
        builder.reset();
        done();
    });

    describe("match", function() {

        it('overwrites an existing match when using the same field again', function (done) {
            builder
                .match("foo", "bar")
                .match("foo", "baz");

            assert.deepEqual(builder._match['foo'], ['baz']);
            done();
        });

        it('adds matches for multiple different fields', function (done) {
            builder
                .match("foo", "baz")
                .match("bar", "baz");

            assert.deepEqual(Object.keys(builder._match), ['foo', 'bar']);
            done();
        });

        it('converts single values to arrays', function (done) {
            builder
                .match("bar", "foo")
                .match("baz", ["foo"]);

            assert.deepEqual(builder._match['bar'], ['foo']);
            assert.deepEqual(builder._match['baz'], ['foo']);
            done();
        });

        it('deletes fields when the value is null or undefined', function (done) {
            builder
                .match("bar", "foo")
                .match("baz", "foo")
                .match("bar", null);

            assert(!('bar' in builder._match));

            builder
                .reset()
                .match("bar", "foo")
                .match("bar");

            assert.equal(builder._match, null);        
            done();
        });

        it('sets _match back to null if last field is deleted', function (done) {
            builder
                .match("bar", "foo")
                .match("bar", null);

            assert.equal(builder._match, null);
            done();
        });

    }); // describe match


    describe("pick", function () {
    
        it('sets this._picked to true when used', function (done) {
            builder
                .pick("foo", "bar");

            assert.equal(builder._picked, true);
            done();
        });

        it('adds picks for multiple fields', function (done) {
            builder
                .pick("bar", "foo")
                .pick("baz", "foo")

            assert.deepEqual(Object.keys(builder._pick), ['bar', 'baz']);
            done();
        });

        it('overwrites an existing match when using the same field again', function (done) {
            builder
                .pick("foo", "bar")
                .pick("foo", "baz");

            assert.equal(builder._pick['foo'], 'baz');
            done();
        });

        it('deletes fields when the value is null or undefined', function (done) {
            builder
                .pick("bar", "foo")
                .pick("baz", "foo")
                .pick("bar", null);

            assert(!('bar' in builder._pick));

            builder
                .reset()
                .pick("bar", "foo")
                .pick("bar");

            assert.equal(builder._pick, null);        
            done();
        });

    }); // describe pick

    
    describe("limit", function () {

        it('is null when the limit is null, undefined or 0', function(done) {
            builder.limit(4);
            assert.equal(builder._limit, 4);  
    
            builder.limit();
            assert.equal(builder._limit, null);  
    
            builder.reset();
            builder
                .limit(4)
                .limit(null);
            assert.equal(builder._limit, null);  
            
            builder.reset();
            builder
                .limit(4)
                .limit(0);
            assert.equal(builder._limit, null);  

            done();
        });

    }); // describe limit

  
    describe("sort", function () {

        it('sets the sort document correctly', function(done) {
            builder.sort("foo", 1);
            assert.deepEqual(builder._sort, {"field": "foo", "direction": 1})

            done();
        });
     
        it('sets sort to null if the field is null or undefined', function(done) {
            builder
                .sort("foo", 1)
                .sort(null);
            assert.equal(builder._sort, null);

            builder
                .reset()
                .sort("foo", 1)
                .sort();
            assert.equal(builder._sort, null);

            done();
        });

    }); // describe sort

    
    describe('schema', function (done) {
        // this test requires running mongoscope and mongodb

        it('waits until the schema is available', function (done) {
            assert.equal(builder._schema, null);
            builder.schema(function (err, res) {
                assert.ifError(err);
                assert.ok(res);
                assert.ok(builder._schema);
                done();
            })

        });

    }); // describe schema


    describe('group', function () {

        it('is null and this._grouped is false before running .group()', function (done) {
            assert.equal(builder._group, null);
            assert.equal(builder._grouped, false);
            done();
        });

        it('sets stores a group and converts field to array', function (done) {
            builder.group("foo", "bar");
            assert.deepEqual(builder._group, {'foo': ['bar']});
            done();
        });

        it('stores compound fields correctly', function (done) {
            builder.group("foo", ["bar", "baz"]);
            assert.deepEqual(builder._group, {'foo': ['bar', 'baz']});
            done();
        });

        it('removes a group name if field is null or undefined', function (done) {
            builder
                .group("foo", ["bar", "baz"])
                .group("foo", null);

            assert.equal(builder._group, null);

            builder
                .reset()
                .group("foo", "bar")
                .group("foo");

            assert.equal(builder._group, null);

            done();
        });

        it('activates nesting only if a second group on a different name is called', function (done) {
            
            builder
                .group("foo", "bar")
                .group("foo", "bar");

            assert.equal(builder._nested, false);

            builder
                .reset()
                .group("foo", "bar")
                .group("baz", "bar");

            assert.equal(builder._nested, true);

            done();
        });

        it('sets _nest_group correctly', function (done) {
            
            builder
                .group("foo", "bar")
                .group("foo", "bar");
            assert.equal(builder._nested, false);
            assert.deepEqual(builder._group, {"foo": ["bar"]});

            builder
                .reset()
                .group("foo", "bar")
                .group("baz", "bar");
            assert.equal(builder._nested, true);
            assert.deepEqual(builder._nest_group, {"foo": ["bar"]});
            assert.deepEqual(builder._group, {"baz": ["bar"]});

            done();
        });

    }); // describe group


    describe("agg", function () {
    
        it('sets this._grouped true when used', function (done) {
            builder
                .agg("foo", "$fn", "bar");

            assert.equal(builder._grouped, true);
            done();
        });

        it('adds aggs for multiple fields', function (done) {
            builder
                .agg("bar", "$fn", "foo")
                .agg("baz", "$fn2", "foo2")

            assert.deepEqual(builder._agg, {"bar": {"$fn": "foo"}, "baz": {"$fn2": "foo2"}});
            done();
        });

        it('overwrites an existing agg when using the same name again', function (done) {
            builder
                .agg("foo", "$fn", "bar")
                .agg("foo", "$fn", "baz");

            assert.deepEqual(builder._agg['foo'], {"$fn": "baz"});
            done();
        });

        it('deletes fields when fn is null or undefined', function (done) {
            builder
                .agg("bar", "$fn", "foo")
                .agg("baz", "$fn", "foo")
                .agg("bar", null);

            assert(!('bar' in builder._agg));

            builder
                .reset()
                .pick("bar", "$fn", "foo")
                .pick("bar");

            assert.equal(builder._pick, null);        
            done();
        });

    }); // describe agg


    describe('_unwinder', function () {

        beforeEach(function(done) {
            // fake schema
            builder._schema = schema;
            done();
        });

        it('undwinds arrays on nested fields without duplication', function(done) {
            builder._unwinder("tags.label");
            builder._unwinder("tags.id");
            assert.deepEqual(builder._unwind, ['tags']);
            done();
        });

        it('undwinds multiply nested arrays', function(done) {
            builder._unwinder("tags.label");
            builder._unwinder("tags.subtags");
            assert.deepEqual(builder._unwind, ['tags', 'tags.subtags']);
            done();
        });

        it('does not unwind non-array fields', function(done) {
            builder._unwinder("user.name.first");
            assert.equal(builder._unwind, null);
            done();
        });

    }); // describe _unwinder


    describe('_compileMatch', function () {

        beforeEach(function(done) {
            // fake schema
            builder._schema = schema;
            done();
        });

        it('returns null if there is nothing to match', function(done) {
            var match_stage = builder._compileMatch();
            assert.equal(match_stage, null);

            builder.reset();
            builder._match = {};
            var match_stage = builder._compileMatch();
            assert.equal(match_stage, null);

            done();
        });

        describe('for type category', function (done) {

            it('matches a single category with an equality match', function(done) {
                builder.match("user.name.last", "Harris");

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, {"$match": {"user.name.last": "Harris"}});
                done();
            });

            it('matches multiple categories with an $in match', function(done) {
                builder.match("user.name.last", ["Harris", "Parker", "Smith"]);

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, {"$match": {"user.name.last": {"$in": ["Harris", "Parker", "Smith"]}}});
                done();
            });

        }); // describe type category

        describe('for type number', function () {
            
            it('matches numbers with $lte / $gte match', function(done) {
                builder.match("user.age", [13, 31]);

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, {"$match": {"user.age": {"$gte": 13, "$lte": 31}}});
                done();
            });

            it('creates an open range on either side if one of the fields is null', function(done) {
                builder.match("user.age", [null, 31]);

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, {"$match": {"user.age": {"$lte": 31}}});

                builder.reset();
                builder.match("user.age", [13, null]);

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, {"$match": {"user.age": {"$gte": 13}}});

                done();
            });

        }); // describe type number


        describe('for type date', function () {
            
            it('matches dates with $lte / $gte match', function(done) {
                builder.match("user.created_at", ["2012-01-01", "2012-12-31"]);

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, 
                    {"$match": {"user.created_at": {"$gte": new Date(Date.parse("2012-01-01")), 
                                                    "$lte": new Date(Date.parse("2012-12-31"))}}});
                done();
            });

            it('creates an open range on either side if one of the fields is null', function(done) {
                builder.match("user.created_at", [null, "2012-12-31"]);

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, {"$match": 
                    {"user.created_at": {"$lte": new Date(Date.parse("2012-12-31"))}}});

                builder.reset();
                builder.match("user.created_at", ["2012-01-01", null]);

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, {"$match": 
                    {"user.created_at": {"$gte": new Date(Date.parse("2012-01-01"))}}});

                done();
            });

            it('accepts Date() objects directly', function (done) {
                builder.match("user.created_at", [new Date(Date.parse("2012-01-01")), null]);

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, 
                    {"$match": {"user.created_at": {"$gte": new Date(Date.parse("2012-01-01"))}}});                

                done();
            });

        }); // describe type date


        describe('for type boolean', function () {

            it('matches booleans with equality matches', function(done) {
                builder.match("user.is_active", true);

                var match_stage = builder._compileMatch();
                assert.deepEqual(match_stage, {"$match": {"user.is_active": true}});

                done();
            });

        });

    }); // describe _compileMatch


    describe('_compileProject', function () {

        it('builds project from group and agg and ignores number fields', function(done) {
            builder
                .group("lastname", "user.name.last")
                    .agg("count", "$sum", 1)
                    .agg("average", "$avg", "user.age");

            var project_stage = builder._compileProject();

            assert.deepEqual(project_stage, {"$project": {"user_name_last": "$user.name.last", "user_age": "$user.age"}});
            done(); 
        });

        it('includes nest_group fields when nested', function(done) {
            builder
                .group("gender", "user.gender")
                    .agg("total", "$sum", "$count")
                .group("lastname", "user.name.last")
                    .agg("count", "$sum", 1)
                    .agg("average", "$avg", "user.age");

            var project_stage = builder._compileProject();

            assert.deepEqual(project_stage, {
                "$project": {
                    "user_gender": "$user.gender",
                    "user_name_last": "$user.name.last", 
                    "user_age": "$user.age"
                }
            });
            done(); 
        });

        it('builds project from pick fields', function(done) {
            builder
                .pick("lastname", "user.name.last")
                .pick("label", "tags.label")
                .pick("zip", "user.address.zip_code")
                .pick("_id", "_id");

            var project_stage = builder._compileProject();

            assert.deepEqual(project_stage, {
                "$project": {
                    "lastname": "$user.name.last", 
                    "label": "$tags.label",
                    "zip": "$user.address.zip_code",
                    "_id": "$_id"
                }
            });


            done();
        });
    }); // describe _compileProject

    
    describe('_compileUnwind', function() {

        beforeEach(function(done) {
            // fake schema
            builder._schema = schema;
            done();
        });

        it('returns null if there is nothing to unwind', function (done) {
            var unwind_stage = builder._compileUnwind();
            assert.equal(unwind_stage, null);

            done();
        });

        it('unwinds a single field correctly', function(done) {

            builder
                .group('labels', 'user.name')
                .agg('ids', '$push', 'tags.id')
                .agg('count', '$sum', 1);

            builder._compileGroup();
            var unwind_stage = builder._compileUnwind();

            assert.deepEqual(unwind_stage, [
                {'$unwind': '$tags'}
            ]);

            done();
        });

        it('unwinds multiple fields correctly', function(done) {

            builder
                .group('labels', 'user.name')
                .agg('subtags', '$push', 'tags.subtags');

            builder._compileGroup();
            var unwind_stage = builder._compileUnwind();

            assert.deepEqual(unwind_stage, [
                {"$unwind": "$tags"},
                {"$unwind": "$tags.subtags"}
            ]);

            done();
        })

    }); // describe _compileUnwind


    describe('_compileGroup', function () {

        beforeEach(function(done) {
            // fake schema
            builder._schema = schema;
            done();
        });

        it('returns null if there is nothing to group', function(done) {
            var group_stage = builder._compileGroup();
            assert.equal(group_stage, null);

            builder.reset();
            builder._group = {};
            var group_stage = builder._compileGroup();
            assert.equal(group_stage, null);

            done();
        });


        it('creates a single group key correctly', function(done) {
            builder
                .group("lastname", "user.name.last");

            var group_stage = builder._compileGroup();
            assert.deepEqual(group_stage, {"$group": {"_id": "$user_name_last"}});

            done();
        });

        it('creates a compound group key correctly', function(done) {
            builder
                .group("fullname", ["user.name.first", "user.name.last"]);

            var group_stage = builder._compileGroup();
            assert.deepEqual(group_stage, 
                {"$group": {"_id": 
                    {"user_name_last": "$user_name_last", 
                     "user_name_first": "$user_name_first"}}});

            done();
        });

        it('groups aggregation fields correctly', function (done) {
            builder
                .group("fullname", ["user.name.first", "user.name.last"])
                .agg("avg_age", "$avg", "user.age")
                .agg("count", "$sum", 1);

            var group_stage = builder._compileGroup();

            assert.deepEqual(group_stage, {
                "$group": {
                    "_id": {
                        "user_name_last": "$user_name_last", 
                        "user_name_first": "$user_name_first"
                    },
                    "avg_age": {
                        "$avg": "$user_age"
                    },
                    "count": {
                        "$sum": 1
                    }
                }
            });

            done();
        })

        it('unwinds nested fields', function(done) {
            
            builder
                .group("labels", ["tags.label"])
                .agg("count", "$sum", 1);

            var group_stage = builder._compileGroup();

            assert.deepEqual(group_stage, {
                "$group": {
                    "_id": "$tags_label",
                    "count": {
                        "$sum": 1
                    }
                }
            });

            assert.deepEqual(builder._unwind, ['tags']);

            done();            
        });
 
    }); // describe _compileGroup


    describe('_compileRename', function() {

        it('returns null if there is no grouping', function (done) {
            var rename_stage = builder._compileRename();
            assert.equal(rename_stage, null);

            done();
        });

        it('renames the group field correctly and removes _id', function(done) {
            
            builder
                .group("fullname", ["user.name.first", "user.name.last"])
                .agg("avg_age", "$avg", "user.age")
                .agg("count", "$sum", 1);

            var rename_stage = builder._compileRename();
            assert.deepEqual(rename_stage, {
                "$project": {
                    "_id": 0,
                    "avg_age": "$avg_age",
                    "count": "$count", 
                    "fullname": "$_id"
                }
            });

            done();
        })

    }); // describe _compileRename


    describe('_compileNest', function () {

        it('returns null if there is no nesting', function (done) {
            var nest_stage = builder._compileNest();
            assert.equal(nest_stage, null);

            done();
        });


        it('builds additional group and project stages', function (done) {
            builder
                .group("gender", "user.gender")
                    .agg("total", "$sum", "count")
                    .group("name", ["user.name.first", "user.name.last"])
                        .agg("avg_age", "$avg", "user.age")
                        .agg("count", "$sum", 1);

            var nest_stage = builder._compileNest();
            
            assert.deepEqual(nest_stage, [
                {
                    "$group": {
                        "_id": "$name.user_gender",
                        "total": {
                            "$sum": "$count"
                        },
                        "values": {
                            "$push": "$$ROOT"
                        }
                    }
                },
                {
                    "$project": {
                        "total": "$total",
                        "gender": "$_id",
                        "_id": 0,
                        "values": "$values"
                    }
                }
            ]);

            done();
        })

    }); // describe _compileNest


    describe('_compileSort', function () {

        it('returns null if there is nothing to unwind', function (done) {
            var sort_stage = builder._compileSort();
            assert.equal(sort_stage, null);

            done();
        });


        it('creates the sort stage correctly', function (done) {
            builder
                .sort("foo", 1);

            var sort_stage = builder._compileSort();
            assert.deepEqual(sort_stage, {'$sort': {'foo': 1}});

            done();
        })

    }); // describe _compileSort


    describe('_compileLimit', function () {

        it('returns null if there is nothing to limit', function (done) {
            var limit_stage = builder._compileLimit();
            assert.equal(limit_stage, null);

            done();
        });


        it('creates the limit stage correctly', function (done) {
            builder
                .limit(42);

            var limit_stage = builder._compileLimit();
            assert.deepEqual(limit_stage, {'$limit': 42});

            done();
        })

    }); // describe _compileLimit


    describe('_compileBin', function () {


        it('bins automatically with auto option', function (done) {

            builder
                .bin('user.created_at', 'auto')
                .group('created', 'user.created_at')

            var bin_stage = builder._compileBin();
            assert.deepEqual(bin_stage[0]['$project']['d'], {'$dayOfMonth': '$user_created_at'});

            done();
        });

        it('looks at match filters to determine date range', function (done) {
            builder
                .match('user.created_at', ['2014-01-01', '2014-03-01'])
                .bin('user.created_at', 'auto')
                .group('created', 'user.created_at')

            var bin_stage = builder._compileBin();
            assert.deepEqual(bin_stage[0]['$project']['d'], {'$dayOfWeek': '$user_created_at'});

            builder
                .reset()
                .match('user.created_at', ['2014-01-01', '2014-01-05'])
                .bin('user.created_at', 'auto')
                .group('created', 'user.created_at')

            var bin_stage = builder._compileBin();
            assert.deepEqual(bin_stage[0]['$project']['d'], undefined);

            done();
        })

        it('includes date binning in pipeline', function (done) {

            builder
                .bin('user.created_at', 'week')
                .group('created', 'user.created_at')
                    .agg('avg_age', '$avg', 'user.age')
                .sort('created', 1)
                .limit(10);

            var pipeline = builder.pipeline();

            assert.deepEqual(pipeline, [
                {
                    "$project": {
                        "user_created_at": "$user.created_at",
                        "user_age": "$user.age"
                    }
                },
                {
                    "$project": {
                        "user_created_at": 1,
                        "user_age": 1,
                        "h": {
                            "$hour": "$user_created_at"
                        },
                        "m": {
                            "$minute": "$user_created_at"
                        },
                        "s": {
                            "$second": "$user_created_at"
                        },
                        "ms": {
                            "$millisecond": "$user_created_at"
                        },
                        "d": {
                            "$dayOfWeek": "$user_created_at"
                        }
                    }
                },
                {
                    "$project": {
                        "user_created_at": {
                            "$subtract": [
                                "$user_created_at",
                                {
                                    "$add": [
                                        "$ms",
                                        {
                                            "$multiply": [
                                                "$s",
                                                1000
                                            ]
                                        },
                                        {
                                            "$multiply": [
                                                "$m",
                                                60,
                                                1000
                                            ]
                                        },
                                        {
                                            "$multiply": [
                                                "$h",
                                                60,
                                                60,
                                                1000
                                            ]
                                        },
                                        {
                                            "$multiply": [
                                                {
                                                    "$subtract": [
                                                        "$d",
                                                        1
                                                    ]
                                                },
                                                24,
                                                60,
                                                60,
                                                1000
                                            ]
                                        }
                                    ]
                                }
                            ]
                        },
                        "user_age": 1
                    }
                },
                {
                    "$group": {
                        "_id": "$user_created_at",
                        "avg_age": {
                            "$avg": "$user_age"
                        }
                    }
                },
                {
                    "$project": {
                        "avg_age": "$avg_age",
                        "created": "$_id",
                        "_id": 0
                    }
                },
                {
                    "$sort": {
                        "created": 1
                    }
                },
                {
                    "$limit": 10
                }
            ]);

            done();
        });

    }); // describe _compileBin


    
    describe('pipeline', function () {

        it('returns a full pipeline for group and agg', function (done) {
            builder
                .group('lastname', 'user.name.last')
                .agg('count', '$sum', 1)
                .sort('count', -1)
                .limit(5);

            var pipeline = builder.pipeline();

            assert.deepEqual(pipeline, [
                {
                    "$project": {
                        "user_name_last": "$user.name.last"
                    }
                },
                {
                    "$group": {
                        "_id": "$user_name_last",
                        "count": { "$sum": 1 }
                    }
                },
                {
                    "$project": {
                        "count":"$count",
                        "lastname":"$_id",
                        "_id": 0
                    }
                },
                {
                    "$sort": {
                        "count": -1
                    }
                },
                {
                    "$limit": 5
                }

            ]);
            done();
        });


        it('includes pipeline stages for nesting', function (done) {
            builder
                .group("gender", ["user.gender", "user.is_active"])
                    .agg("total", "$sum", "count")
                    .group("name", ["user.name.first", "user.name.last"])
                        .agg("avg_age", "$avg", "user.age")
                        .agg("count", "$sum", 1);

            var pipeline = builder.pipeline();
            console.log(pipeline);
            done();
        });

    }); // describe pipeline

    
    describe('end', function () {

        it('throws and error when both group and pick is used', function (done) {
            assert.throws(function () {
                builder
                    .group("foo", "bar")
                    .pick("foo", "bar")
                    .end(function () {});
                },
                /pick and group/
            );

            done();
        });

        it('returns a result', function (done) {
            // this test requires running mongoscope and mongodb
           
            builder
                .group('lastname', 'user.name.last')
                .agg('count', '$sum', 1)
                .sort('count', -1)
                .limit(5);

            assert.doesNotThrow(function () {
                builder.end(function () {});
            });

            builder.end(function (err, res) {
                assert.ifError(err);
                assert.ok(res);
                done();            
            });
        });

    }); // describe end

}); // describe QueryBuilder











